CREATE TABLE catalog.db.intervals_spec (
  interval_year INTERVAL YEAR,
  interval_month INTERVAL MONTH,
  interval_year_to_month INTERVAL YEAR TO MONTH,
  interval_day INTERVAL DAY,
  interval_hour INTERVAL HOUR,
  interval_minute INTERVAL MINUTE,
  interval_second INTERVAL SECOND,
  interval_day_to_hour INTERVAL DAY TO HOUR
); 